#include <iterator>
#include <exception>
template <typename T>
   struct list_t
{
   typedef T value_type;
   typedef T& reference;
   typedef T const & const_reference;
   typedef T* pointer;
   typedef T const * const_pointer;
   typedef ptrdiff_t difference_type;
   typedef size_t size_type;

   struct const_iterator;
   struct iterator;

   typedef std::reverse_iterator<iterator> reverse_iterator;
   typedef std::reverse_iterator<const_iterator> const_reverse_iterator;

   list_t();
   template <typename Iter>
      list_t(Iter begin, Iter end);
   list_t(list_t const &);
   list_t &operator =(list_t a)
   {
      swap(a);
      return *this;
   }

   iterator begin();
   iterator end();
   const_iterator begin() const;
   const_iterator end() const;

   reverse_iterator rbegin();
   reverse_iterator rend();
   const_reverse_iterator rbegin() const;
   const_reverse_iterator rend() const;


   template <typename Iter>
      void assign(Iter begin, Iter end);

   reference front();

   reference back();

   void push_back(T const &);

   void push_front(T const &);

   void pop_front();

   void pop_back();

   void swap(list_t &);

   iterator erase(iterator);
   iterator insert(iterator, value_type const&);

   void clear();
   bool empty() const;
   ~list_t();

   
private:
   struct element;
   element *head_;
   element *tail_;

};

   template <typename T>
      template <typename Iter>
         void list_t<T>::assign( Iter begin, Iter end )
   {
      list_t<T> tmp(begin, end);
      swap(tmp);
   }

   template <typename T>
      typename list_t<T>::reverse_iterator list_t<T>::rend()
   {
      return reverse_iterator(begin());
   }

   template <typename T>
      typename list_t<T>::reverse_iterator list_t<T>::rbegin()
   {
      return reverse_iterator(end());
   }

   template <typename T>
      typename list_t<T>::const_reverse_iterator list_t<T>::rend() const
   {
      return const_reverse_iterator(begin());
   }




   template <typename T>
      typename list_t<T>::const_reverse_iterator list_t<T>::rbegin() const
   {
      return const_reverse_iterator(end());
   }

   template <typename T>
   list_t<T>::~list_t()
   {
      clear();
      delete head_;
   }

   template <typename T>
   void list_t<T>::clear()
   {
      while (!empty())
      {
         pop_back();
      }
   }

   template <typename T>
   bool list_t<T>::empty() const
   {
      return head_ == tail_;
   }



   template <typename T>
   struct list_t<T>::const_iterator 
   {
      friend struct list_t<T>;
      typedef std::bidirectional_iterator_tag iterator_category;
      typedef  T value_type;
      typedef ptrdiff_t  difference_type;
      typedef T const *pointer;
      typedef T const &reference;



      const_iterator(element *const ptr)
         :ptr_(ptr)
      { }
      const_iterator operator++(int a)
      {
         *element tmp = this;
         operator++;
         return tmp;
      }
      const_iterator operator--(int a)
      {
         *element tmp = this;
         operator--;
         return tmp;
      }
      const_iterator & operator++()
      {
         ptr_ = ptr_->next_;
         return *this;
      }
      const_iterator & operator--()
      {
         ptr_ = ptr_->prev_;
         return *this;
      }
      T const &operator*() const
      {
         return ptr_->value_;
      }

   
      bool operator!=(const_iterator& a)
      {
         return a.ptr_ != ptr_;
      }
      bool operator==(const_iterator& a)
      {
         return a.ptr_ == ptr_;
      }
   private:
      element const *get_ref()
      {
         return ptr_;
      }

   protected:
      element *ptr_;
   };

   template <typename T>
   struct list_t<T>::iterator: const_iterator 
   {
      friend struct list_t<T>;
      typedef T *pointer;
      typedef T &reference;
      iterator(element *const ptr)
         :const_iterator(ptr)
      { }
      T &operator*()
      {
         return ptr_->value_;
      }
   private:
      element *get_ref()
      {
         return ptr_;
      }
   };

   template <typename T>
   struct list_t<T>::element
   {
      element()
         :value_(T()), next_(NULL), prev_(NULL)
      { }
      element(T const & value, element *const next, element *const prev)
         :value_(value), next_(next), prev_(prev)
      {}

      T value_;
      element *next_;
      element *prev_;

   };


template <typename T>
   typename list_t<T>::iterator list_t<T>::insert( iterator iter, value_type const& a)
{
   element *tmp = iter.get_ref();
   if (head_ == tmp)
   {
      push_front(a);
      return iterator(head_);
   }
   element *tmp_element = new element(a, tmp, tmp->prev_);
   tmp->prev_->next_ = tmp_element;
   tmp->prev_ = tmp_element;
}



template <typename T>
   typename list_t<T>::iterator list_t<T>::erase( iterator iter )
{
   element *tmp = iter.get_ref();
   iterator ret(tmp->next_);
   if (tmp->prev_)
   {
      tmp->prev_->next_ = tmp->next_;
      tmp->next_->prev_ = tmp->prev_;
      return ret;
   }
   pop_front();
   return ret;
}

template <typename T>
   typename list_t<T>::const_iterator list_t<T>::end() const
{
   return const_iterator(tail_);
}

template <typename T>
   typename list_t<T>::const_iterator list_t<T>::begin() const
{
   return const_iterator(head_);
}

template <typename T>
   typename list_t<T>::iterator list_t<T>::end()
{
   return iterator(tail_);
}

template <typename T>
   typename list_t<T>::iterator list_t<T>::begin()
{
   return iterator(head_);
}

template <typename T>
template <typename Iter>
   list_t<T>::list_t(Iter begin, Iter end)
{
   element *fake = new element;
   head_ = fake;
   tail_ = fake;
   for (Iter i = begin, i_end = end; i != i_end; ++i)
   {
      push_back(*i);
   }
}

template <typename T>
   list_t<T>::list_t( list_t const &a )
{
   element *fake = new element;
   head_ = fake;
   tail_ = fake;
   for (list_t<T>::const_iterator i = a.begin(), i_end = a.end(); i != i_end; ++i)
   {
      push_back(*i);
   }
}

template <typename T>
   void list_t<T>::swap( list_t &a )
{
   std::swap(head_, a.head_);
   std::swap(tail_, a.tail_);
}

template <typename T>
   void list_t<T>::pop_back()
{
   if (head_ == tail_)
   {
      throw std::runtime_error("Pop from empty list");
   }
   element * tmp = tail_->prev_;
   if (tmp->prev_)
   {
      tmp->prev_->next_ = tail_;
      tail_->prev_ = tmp->prev_;
   } 
   else
   {
      head_ = tail_;
   }
   delete tmp;
}

template <typename T>
   void list_t<T>::pop_front()
{
   if (head_ == tail_)
   {
      throw std::runtime_error("Pop from empty list");
   }
   element * tmp = head_;
   head_->next_->prev_= NULL;
   head_ = head_->next_;
   delete tmp;
}

template <typename T>
   void list_t<T>::push_front( T const & a)
{
   element *tmp = new element(a, head_, NULL);
   head_->prev_ = tmp;
   head_ = tmp;
}

template <typename T>
   void list_t<T>::push_back( T const & a )
{
   if (head_ == tail_)
   {
      push_front(a);
      return;
   }
   element *tmp = new element(a, tail_, tail_->prev_);
   tail_->prev_->next_ = tmp;
   tail_->prev_ = tmp;
}

template <typename T>
   list_t<T>::list_t()
{
   element *fake = new element;
   head_ = fake;
   tail_ = fake;
}

template <typename T>
   typename list_t<T>::reference list_t<T>::back()
{
   if (head_ == tail_)
   {
      throw std::runtime_error("Back form empty list");
   }
   return tail_->prev_->value_;
}

template <typename T>
   typename list_t<T>::reference list_t<T>::front()
{
   if (head_ == tail_)
   {
      throw std::runtime_error("Front from empty list");
   }
   return head_->value_;
}

template <typename T>
   int compare(list_t<T> const &a, list_t<T> const &b)
{
   typename list_t<T>::const_iterator i = a.begin(), i_end = a.end(), j = b.begin(), j_end = b.end();
   while (i != i_end)
   {
      if (j == j_end)
      {
         return 1;
      }
      if (*i < *j)
      {
         return -1;
      }

      if (*i > *j)
      {
         return 1;
      }
      ++i;
      ++j;
   }
   if (j != j_end)
   {
      return -1;
   }
   return 0;
}
template <typename T>
bool operator ==(list_t<T> const &a, list_t<T> const &b)
{
   return compare(a, b) == 0;
}

template <typename T>
bool operator !=(list_t<T> const &a, list_t<T> const &b)
{
   return compare(a, b) != 0;
}

template <typename T>
bool operator <=(list_t<T> const &a, list_t<T> const &b)
{
   return compare(a, b) <= 0;
}

template <typename T>
bool operator >=(list_t<T> const &a, list_t<T> const &b)
{
   return compare(a, b) >= 0;
}

template <typename T>
bool operator <(list_t<T> const &a, list_t<T> const &b)
{
   return compare(a, b) < 0;
}

template <typename T>
bool operator >(list_t<T> const &a, list_t<T> const &b)
{
   return compare(a, b) > 0;
}